# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Grid Search Designer which searches over a discretized grid of Trial parameter values."""

import copy
import random
from typing import Dict, List, Optional, Sequence

from absl import logging
import attrs
import numpy as np
from vizier import algorithms
from vizier import pyvizier
from vizier.interfaces import serializable
from vizier.pyvizier import converters


GridValues = Dict[str, List[pyvizier.ParameterValue]]


@attrs.define(auto_attribs=False, init=False)
class GridSearchDesigner(algorithms.PartiallySerializableDesigner):
  """Grid Search designer.

  This designer searches over a grid of hyper-parameter values.

  NOTE: The grid search index (i.e. which grid point to output) is based the
  number of suggestions created so far (regardless of completion or not). This
  means the class must be wrapped via `PartiallySerializableDesignerPolicy` for
  use in Pythia, thus requiring load/dump implementations.
  """

  _unshuffled_grid_values: GridValues = attrs.field()
  _grid_values: GridValues = attrs.field()
  _current_index: int = attrs.field()
  _shuffle_seed: Optional[int] = attrs.field()
  _double_grid_resolution: int = attrs.field()

  _metadata_ns: str = 'grid'  # class-level constant.

  def __init__(
      self,
      search_space: pyvizier.SearchSpace,
      shuffle_seed: Optional[int] = None,
      *,
      double_grid_resolution: int = 10,
  ):
    """Init.

    Args:
      search_space: Must be a flat search space.
      shuffle_seed: Whether to shuffle the grid ordering. If None, uses the
        given ordering from original search space.
      double_grid_resolution: Number of grid points for DOUBLE parameters.
    """
    if search_space.is_conditional:
      raise ValueError(
          f'This designer {self} does not support conditional search.'
      )
    self._search_space = search_space
    self._shuffle_seed = shuffle_seed
    self._double_grid_resolution = double_grid_resolution
    self._current_index = 0

    # Creates unshuffled grid values for every parameter. This is just a
    # template for creating the potentially shuffled self._grid_values to be
    # used in suggest().
    self._unshuffled_grid_values = {}
    for parameter_config in self._search_space.parameters:
      self._unshuffled_grid_values[parameter_config.name] = (
          self._grid_points_from_parameter_config(parameter_config)
      )

    # Set true grid values to be used during suggest calls.
    self._grid_values = self._maybe_shuffled_grid_values(self._shuffle_seed)

  @classmethod
  def from_problem(
      cls,
      problem: pyvizier.ProblemStatement,
      seed: Optional[int] = None,
  ):
    """For wrapping via `PartiallySerializableDesignerPolicy`."""
    return GridSearchDesigner(problem.search_space, seed)

  def update(
      self,
      completed: algorithms.CompletedTrials,
      all_active: algorithms.ActiveTrials,
  ) -> None:
    pass

  def suggest(
      self, count: Optional[int] = None
  ) -> Sequence[pyvizier.TrialSuggestion]:
    """Make new suggestions.

    Args:
      count: Makes best effort to generate this many suggestions. If None,
        suggests as many as the algorithm wants.

    Returns:
      New suggestions.
    """
    count = count or 1

    parameter_dicts = []
    for index in range(self._current_index, self._current_index + count):
      # Use index to select parameters via Cartesian Product ordering.
      # Effectively equivalent to itertools.product(list_of_lists)[index]`,
      # without the memory blowup.
      parameter_dict = pyvizier.ParameterDict()
      temp_index = index
      for p_name in self._grid_values:
        p_length = len(self._grid_values[p_name])
        p_index = temp_index % p_length
        parameter_dict[p_name] = self._grid_values[p_name][p_index]
        temp_index = temp_index // p_length
      parameter_dicts.append(parameter_dict)

    self._current_index += len(parameter_dicts)
    return [pyvizier.TrialSuggestion(parameters=p) for p in parameter_dicts]

  def load(self, metadata: pyvizier.Metadata) -> None:
    """Load the current index."""
    metadata = metadata.ns(self._metadata_ns)
    try:
      current_index = int(metadata['current_index'])
      none_or_int = metadata['shuffle_seed']
      shuffle_seed = None if (none_or_int == 'None') else int(none_or_int)
      logging.info('Restored shuffle seed: %s', shuffle_seed)
    except (KeyError, ValueError) as e:
      raise serializable.HarmlessDecodeError() from e

    self._current_index = current_index
    self._shuffle_seed = shuffle_seed
    self._grid_values = self._maybe_shuffled_grid_values(self._shuffle_seed)

  def dump(self) -> pyvizier.Metadata:
    """Dump the current index."""
    metadata = pyvizier.Metadata()
    metadata.ns(self._metadata_ns)['current_index'] = str(self._current_index)
    metadata.ns(self._metadata_ns)['shuffle_seed'] = str(self._shuffle_seed)
    return metadata

  def _grid_points_from_parameter_config(
      self,
      parameter_config: pyvizier.ParameterConfig,
  ) -> List[pyvizier.ParameterValue]:
    """Produces grid points from a parameter_config."""
    if parameter_config.type == pyvizier.ParameterType.DOUBLE:
      min_value, max_value = parameter_config.bounds
      if min_value == max_value:
        return [pyvizier.ParameterValue(value=min_value)]

      converter = converters.DefaultModelInputConverter(
          parameter_config, scale=True
      )
      grid_scalars = np.linspace(0.0, 1.0, num=self._double_grid_resolution)
      return converter.to_parameter_values(grid_scalars)  # pytype:disable=bad-return-type

    elif parameter_config.type == pyvizier.ParameterType.INTEGER:
      min_value, max_value = parameter_config.bounds
      return [
          pyvizier.ParameterValue(value=value)
          for value in range(min_value, max_value + 1)
      ]

    elif parameter_config.type in [
        pyvizier.ParameterType.DISCRETE,
        pyvizier.ParameterType.CATEGORICAL,
    ]:
      return [
          pyvizier.ParameterValue(value=value)
          for value in parameter_config.feasible_values
      ]

    else:
      raise ValueError(
          'ParameterConfig type is not one of the supported primitives for'
          f' ParameterConfig: {parameter_config}'
      )

  def _maybe_shuffled_grid_values(
      self, shuffle_seed: Optional[int]
  ) -> GridValues:
    grid_values = copy.deepcopy(self._unshuffled_grid_values)

    # Shuffle the grid if specified.
    if shuffle_seed is not None:
      rng = random.Random(shuffle_seed)
      # Shuffle dict keys.
      shuffled_items = list(grid_values.items())
      rng.shuffle(shuffled_items)
      grid_values = dict(shuffled_items)
      # Shuffle dict value lists.
      for param_name in grid_values:
        rng.shuffle(grid_values[param_name])

    return grid_values



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""EmukitDesigner wraps emukit into Vizier Designer."""

import enum
from typing import Optional, Sequence

from absl import logging
from emukit import core
from emukit import model_wrappers
from emukit.bayesian_optimization import acquisitions
from emukit.bayesian_optimization import loops
from emukit.core import initial_designs
from GPy import kern
from GPy import models
from GPy.core.parameterization import priors
from GPy.util import linalg
import numpy as np
from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier.pyvizier import converters

RandomDesign = initial_designs.RandomDesign


def _create_constrained_gp(features: np.ndarray, labels: np.ndarray):
  """Creates a constraint gp."""

  # This logging is too chatty because paramz transformations do not implement
  # log jacobians. Silence it.
  logging.logging.getLogger('paramz.transformations').setLevel(
      logging.logging.CRITICAL)

  class LogGaussian:
    """Multi-variate version of Log-Gaussian.

    GPy surprisingly doesn't have this. The expected API of lnpdf and lnpdf_grad
    are not precisely defined, so this handwaves a lot of stuff based on how
    MultiVariateGaussian is implemented.
    """
    domain = 'positive'

    def __init__(self, mu, sigma):
      self.mu = (mu)
      self.sigma = (sigma)
      self.inv, _, self.hld, _ = linalg.pdinv(self.sigma)
      self.sigma2 = np.square(self.sigma)
      self.constant = -0.5 * (self.mu.size * np.log(2 * np.pi) + self.hld)

    def lnpdf(self, x):
      x = np.array(x).flatten()
      d = np.log(x) - self.mu
      # Constant is dropped. Exact value doesn't really matter. Hopefully.
      return -0.5 * np.dot(d.T, np.dot(self.inv, d))

    def lnpdf_grad(self, x):
      x = np.array(x).flatten()
      d = np.log(x) - self.mu
      return -np.dot(self.inv, d)

    def rvs(self, n):
      return np.exp(
          np.random.randn(int(n), self.sigma.shape[0]) * self.sigma + self.mu)

  # Use heavy tailed priors, but start with small values.
  kernel = kern.Matern52(features.shape[1], variance=.04, ARD=True)
  kernel.unconstrain()
  loggaussian = LogGaussian(
      np.zeros(features.shape[1:]),
      sigma=np.diag(np.ones(features.shape[1:]) * 4.6))
  kernel.lengthscale.set_prior(loggaussian)
  kernel.lengthscale.constrain_bounded(1e-2, 1e2)
  kernel.variance.set_prior(priors.LogGaussian(-3.2, 4.6))
  kernel.variance.constrain_bounded(1e-3, 1e1)

  gpy_model = models.GPRegression(features, labels, kernel, noise_var=0.0039)
  gpy_model.likelihood.unconstrain()
  gpy_model.likelihood.variance.set_prior(priors.LogGaussian(-5.5, sigma=4.6))
  gpy_model.likelihood.variance.constrain_bounded(1e-10, 1.)

  gpy_model.optimize_restarts(20, robust=True, optimizer='lbfgsb')
  logging.info('After train: %s, %s', gpy_model, gpy_model.kern.lengthscale)
  return gpy_model


def _to_emukit_parameter(spec: converters.NumpyArraySpec) -> core.Parameter:
  if spec.type == converters.NumpyArraySpecType.ONEHOT_EMBEDDING:
    return core.CategoricalParameter(
        spec.name,
        core.OneHotEncoding(list(range(spec.num_dimensions - spec.num_oovs))))
  elif spec.type == converters.NumpyArraySpecType.CONTINUOUS:
    return core.ContinuousParameter(spec.name, spec.bounds[0], spec.bounds[1])
  else:
    raise ValueError(f'Unknown type: {spec.type.name} in {spec}')


def _to_emukit_parameters(search_space: vz.SearchSpace) -> core.ParameterSpace:
  parameters = [_to_emukit_parameter(pc) for pc in search_space.parameters]  # pytype: disable=wrong-arg-types  # always-use-return-annotations
  return core.ParameterSpace(parameters)


class Version(enum.Enum):
  DEFAULT_EI = 'emukit_default_ei'
  MATERN52_UCB = 'emukit_matern52_ucb_ard'


class EmukitDesigner(vza.Designer):
  """Wraps emukit library as a Vizier designer."""

  def __init__(self,
               study_config: vz.ProblemStatement,
               *,
               version: Version = Version.DEFAULT_EI,
               num_random_samples: int = 10,
               metadata_ns: str = 'emukit'):
    """Init.

    Args:
      study_config: Must be a flat study with a single metric.
      version: Determines the behavior. See Version.
      num_random_samples: This designer suggests random points until this many
        trials have been observed.
      metadata_ns: Metadata namespace that this designer writes to.

    Raises:
      ValueError:
    """
    if study_config.search_space.is_conditional:
      raise ValueError(f'{type(self)} does not support conditional search.')
    if len(study_config.metric_information) != 1:
      raise ValueError(f'{type(self)} works with exactly one metric.')
    self._study_config = study_config
    self._version = Version(version)

    # Emukit pipeline's model and acquisition optimizer use the same
    # representation. We need to remove the oov dimensions.
    self._converter = converters.TrialToArrayConverter.from_study_config(
        study_config,
        scale=True,
        flip_sign_for_minimization_metrics=True,
        pad_oovs=False,
    )
    self._trials = tuple()
    self._emukit_space = core.ParameterSpace(
        [_to_emukit_parameter(spec) for spec in self._converter.output_specs])

    self._metadata_ns = metadata_ns
    self._num_random_samples = num_random_samples

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    self._trials += tuple(completed.trials)

  def _suggest_random(self, count: int) -> Sequence[vz.TrialSuggestion]:
    sampler = RandomDesign(self._emukit_space)  # Collect random points
    samples = sampler.get_samples(count)
    return self._to_suggestions(samples, 'random')

  def _to_suggestions(self, arr: np.ndarray,
                      source: str) -> Sequence[vz.TrialSuggestion]:
    """Convert arrays to suggestions and record the source."""
    suggestions = []
    for params in self._converter.to_parameters(arr):
      suggestion = vz.TrialSuggestion(params)
      suggestion.metadata.ns(self._metadata_ns)['source'] = source
      suggestions.append(suggestion)
    return suggestions

  def _suggest_bayesopt(self,
                        count: Optional[int] = None
                       ) -> Sequence[vz.TrialSuggestion]:
    features, labels = self._converter.to_xy(self._trials)
    # emukit minimizes. Flip the signs again (converter flips for minimization).
    labels = -labels

    if self._version == Version.DEFAULT_EI:
      gpy_model = models.GPRegression(features, labels)
      emukit_model = model_wrappers.GPyModelWrapper(gpy_model)
      acquisition = acquisitions.ExpectedImprovement(model=emukit_model)

    elif self._version == Version.MATERN52_UCB:
      gpy_model = _create_constrained_gp(features, labels)
      emukit_model = model_wrappers.GPyModelWrapper(gpy_model, n_restarts=20)
      acquisition = acquisitions.NegativeLowerConfidenceBound(  # pytype: disable=wrong-arg-types  # numpy-scalars
          model=emukit_model, beta=np.float64(1.8)
      )

    logging.info(
        'Emukit model: model=%s',
        # Values associated with length-1 keys are useless.
        {k: v for k, v in emukit_model.model.to_dict().items() if len(k) > 1})

    logging.info('Gpy model: model=%s', gpy_model)

    bayesopt_loop = loops.BayesianOptimizationLoop(
        model=emukit_model,
        space=self._emukit_space,
        acquisition=acquisition,
        batch_size=count or 1)
    x1 = bayesopt_loop.get_next_points([])

    return self._to_suggestions(x1, 'bayesopt')

  def suggest(self,
              count: Optional[int] = None) -> Sequence[vz.TrialSuggestion]:
    if len(self._trials) < self._num_random_samples:
      return self._suggest_random(
          count or (self._num_random_samples - len(self._trials)))

    try:
      return self._suggest_bayesopt(count)
    except np.linalg.LinAlgError as e:
      logging.exception('Training failed: %s', e)
      return self._suggest_random(count or 1)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Converters for OSS Vizier's protos from/to PyVizier's classes."""

import datetime
import logging
from typing import Iterable, List, Optional, Sequence, Tuple, Union

from absl import logging
from vizier._src.pythia import policy
from vizier._src.pyvizier.oss import metadata_util
from vizier._src.pyvizier.pythia import study
from vizier._src.pyvizier.shared import base_study_config
from vizier._src.pyvizier.shared import common
from vizier._src.pyvizier.shared import parameter_config
from vizier._src.pyvizier.shared import trial
from vizier._src.service import pythia_service_pb2
from vizier._src.service import study_pb2
from vizier._src.service import vizier_service_pb2


ScaleType = parameter_config.ScaleType
_ScaleTypePb2 = study_pb2.StudySpec.ParameterSpec.ScaleType
ParameterType = parameter_config.ParameterType
MonotypeParameterSequence = parameter_config.MonotypeParameterSequence


class StudyStateConverter:
  """Proto converter for Study states."""

  _pyvizier_to_proto = {
      study.StudyState.ACTIVE: study_pb2.Study.State.ACTIVE,
      study.StudyState.ABORTED: study_pb2.Study.State.INACTIVE,
      study.StudyState.COMPLETED: study_pb2.Study.State.COMPLETED,
  }
  _proto_to_pyvizier = {v: k for k, v in _pyvizier_to_proto.items()}

  @classmethod
  def to_proto(cls, state: study.StudyState) -> study_pb2.Study.State:
    if state in cls._pyvizier_to_proto:
      return cls._pyvizier_to_proto[state]
    return study_pb2.Study.State.STATE_UNSPECIFIED

  @classmethod
  def from_proto(cls, proto: study_pb2.Study.State) -> study.StudyState:
    if proto in cls._proto_to_pyvizier:
      return cls._proto_to_pyvizier[proto]
    elif proto == study_pb2.Study.State.STATE_UNSPECIFIED:
      # OSS Vizier server treats STATE_UNSPECIFIED as ACTIVE.
      return study.StudyState.ACTIVE
    else:
      raise ValueError(
          'Proto Study state {} has no equivalent in PyVizier.'.format(
              study_pb2.Study.State.Name(proto)
          )
      )


class _ScaleTypeMap:
  """Proto converter for scale type."""

  _pyvizier_to_proto = {
      ScaleType.LINEAR: _ScaleTypePb2.UNIT_LINEAR_SCALE,
      ScaleType.LOG: _ScaleTypePb2.UNIT_LOG_SCALE,
      ScaleType.REVERSE_LOG: _ScaleTypePb2.UNIT_REVERSE_LOG_SCALE,
  }
  _proto_to_pyvizier = {v: k for k, v in _pyvizier_to_proto.items()}

  @classmethod
  def to_proto(cls, pyvizier: ScaleType) -> _ScaleTypePb2:
    return cls._pyvizier_to_proto[pyvizier]

  @classmethod
  def from_proto(cls, proto: _ScaleTypePb2) -> ScaleType:
    return cls._proto_to_pyvizier[proto]


class ParameterConfigConverter:
  """Converter for ParameterConfig."""

  @classmethod
  def _set_bounds(
      cls,
      proto: study_pb2.StudySpec.ParameterSpec,
      lower: float,
      upper: float,
      parameter_type: ParameterType,
  ):
    """Sets the proto's min_value and max_value fields."""
    if parameter_type == ParameterType.INTEGER:
      proto.integer_value_spec.min_value = lower
      proto.integer_value_spec.max_value = upper
    elif parameter_type == ParameterType.DOUBLE:
      proto.double_value_spec.min_value = lower
      proto.double_value_spec.max_value = upper

  @classmethod
  def _set_feasible_points(
      cls,
      proto: study_pb2.StudySpec.ParameterSpec,
      feasible_points: Sequence[float],
  ):
    """Sets the proto's feasible_points field."""
    feasible_points = sorted(feasible_points)
    proto.discrete_value_spec.ClearField('values')
    proto.discrete_value_spec.values.extend(feasible_points)

  @classmethod
  def _set_categories(
      cls, proto: study_pb2.StudySpec.ParameterSpec, categories: Sequence[str]
  ):
    """Sets the protos' categories field."""
    proto.categorical_value_spec.ClearField('values')
    proto.categorical_value_spec.values.extend(categories)

  @classmethod
  def _set_default_value(
      cls,
      proto: study_pb2.StudySpec.ParameterSpec,
      default_value: Union[float, int, str],
  ):
    """Sets the protos' default_value field."""
    which_pv_spec = proto.WhichOneof('parameter_value_spec')
    getattr(proto, which_pv_spec).default_value.value = default_value

  @classmethod
  def _matching_parent_values(
      cls, proto: study_pb2.StudySpec.ParameterSpec.ConditionalParameterSpec
  ) -> MonotypeParameterSequence:
    """Returns the matching parent values, if set."""
    oneof_name = proto.WhichOneof('parent_value_condition')
    if not oneof_name:
      return []
    if oneof_name in (
        'parent_discrete_values',
        'parent_int_values',
        'parent_categorical_values',
    ):
      return list(getattr(getattr(proto, oneof_name), 'values'))
    raise ValueError('Unknown matching_parent_vals: {}'.format(oneof_name))

  @classmethod
  def from_proto(
      cls,
      proto: study_pb2.StudySpec.ParameterSpec,
      *,
      strict_validation: bool = False,
  ) -> parameter_config.ParameterConfig:
    """Creates a ParameterConfig.

    Args:
      proto:
      strict_validation: If True, raise ValueError to enforce that
        from_proto(proto).to_proto == proto.

    Returns:
      ParameterConfig object

    Raises:
      ValueError: See the "strict_validtion" arg documentation.
    """
    feasible_values = []
    oneof_name = proto.WhichOneof('parameter_value_spec')
    if oneof_name == 'integer_value_spec':
      bounds = (
          int(proto.integer_value_spec.min_value),
          int(proto.integer_value_spec.max_value),
      )
    elif oneof_name == 'double_value_spec':
      bounds = (
          proto.double_value_spec.min_value,
          proto.double_value_spec.max_value,
      )
    elif oneof_name == 'discrete_value_spec':
      bounds = None
      feasible_values = proto.discrete_value_spec.values
    elif oneof_name == 'categorical_value_spec':
      bounds = None
      feasible_values = proto.categorical_value_spec.values

    default_value = None
    if getattr(proto, oneof_name).default_value.value:
      default_value = getattr(proto, oneof_name).default_value.value

    if proto.conditional_parameter_specs:
      children = []
      for conditional_ps in proto.conditional_parameter_specs:
        parent_values = cls._matching_parent_values(conditional_ps)
        children.append(
            (parent_values, cls.from_proto(conditional_ps.parameter_spec))
        )
    else:
      children = None

    scale_type = None
    if proto.scale_type:
      scale_type = _ScaleTypeMap.from_proto(proto.scale_type)

    try:
      config = parameter_config.ParameterConfig.factory(
          name=proto.parameter_id,
          feasible_values=feasible_values,
          bounds=bounds,
          children=children,
          scale_type=scale_type,
          default_value=default_value,
      )
    except ValueError as e:
      raise ValueError(
          'The provided proto was misconfigured. {}'.format(proto)
      ) from e

    if strict_validation and cls.to_proto(config) != proto:
      raise ValueError(
          'The provided proto was misconfigured. Expected: {} Given: {}'.format(
              cls.to_proto(config), proto
          )
      )
    return config

  @classmethod
  def _set_child_parameter_configs(
      cls,
      parent_proto: study_pb2.StudySpec.ParameterSpec,
      pc: parameter_config.ParameterConfig,
  ):
    """Sets the parent_proto's conditional_parameter_specs field.

    Args:
      parent_proto: Modified in place.
      pc: Parent ParameterConfig to copy children from.

    Raises:
      ValueError: If the child configs are invalid
    """
    children: List[
        Tuple[MonotypeParameterSequence, parameter_config.ParameterConfig]
    ] = []
    for child in pc.child_parameter_configs:
      children.append((child.matching_parent_values, child))
    if not children:
      return

    parent_proto.ClearField('conditional_parameter_specs')
    for child_pair in children:
      if len(child_pair) != 2:
        raise ValueError(
            """Each element in children must be a tuple of
            (Sequence of valid parent values,  ParameterConfig)"""
        )

    logging.debug(
        '_set_child_parameter_configs: parent_proto=%s, children=%s',
        parent_proto,
        children,
    )
    for unsorted_parent_values, child in children:
      parent_values = sorted(unsorted_parent_values)
      child_proto = cls.to_proto(child.clone_without_children)
      conditional_parameter_spec = (
          study_pb2.StudySpec.ParameterSpec.ConditionalParameterSpec(
              parameter_spec=child_proto
          )
      )

      if parent_proto.HasField('discrete_value_spec'):
        conditional_parameter_spec.parent_discrete_values.values[:] = (
            parent_values
        )
      elif parent_proto.HasField('categorical_value_spec'):
        conditional_parameter_spec.parent_categorical_values.values[:] = (
            parent_values
        )
      elif parent_proto.HasField('integer_value_spec'):
        conditional_parameter_spec.parent_int_values.values[:] = parent_values
      else:
        raise ValueError('DOUBLE type cannot have child parameters')
      if child.child_parameter_configs:
        cls._set_child_parameter_configs(child_proto, child)
      parent_proto.conditional_parameter_specs.extend(
          [conditional_parameter_spec]
      )

  @classmethod
  def to_proto(
      cls, pc: parameter_config.ParameterConfig
  ) -> study_pb2.StudySpec.ParameterSpec:
    """Returns a ParameterConfig Proto."""
    proto = study_pb2.StudySpec.ParameterSpec(parameter_id=pc.name)
    if pc.type == ParameterType.DISCRETE:
      cls._set_feasible_points(proto, [float(v) for v in pc.feasible_values])
    elif pc.type == ParameterType.CATEGORICAL:
      cls._set_categories(proto, pc.feasible_values)
    elif pc.type in (ParameterType.INTEGER, ParameterType.DOUBLE):
      cls._set_bounds(proto, pc.bounds[0], pc.bounds[1], pc.type)
    else:
      raise ValueError('Invalid ParameterConfig: {}'.format(pc))
    if (
        pc.scale_type is not None
        and pc.scale_type != ScaleType.UNIFORM_DISCRETE
    ):
      proto.scale_type = _ScaleTypeMap.to_proto(pc.scale_type)
    if pc.default_value is not None:
      cls._set_default_value(proto, pc.default_value)

    cls._set_child_parameter_configs(proto, pc)
    return proto


class ParameterValueConverter:
  """Converter for vz.ParameterValue."""

  @classmethod
  def from_proto(
      cls, proto: study_pb2.Trial.Parameter
  ) -> Optional[trial.ParameterValue]:
    """Returns whichever value that is populated, or None."""
    value_proto = proto.value
    oneof_name = value_proto.WhichOneof('kind')
    potential_value = getattr(value_proto, oneof_name)
    if (
        isinstance(potential_value, float)
        or isinstance(potential_value, str)
        or isinstance(potential_value, bool)
    ):
      return trial.ParameterValue(potential_value)
    else:
      return None

  @classmethod
  def to_proto(
      cls, parameter_value: trial.ParameterValue, name: str
  ) -> study_pb2.Trial.Parameter:
    """Returns Parameter Proto."""
    proto = study_pb2.Trial.Parameter(parameter_id=name)

    if isinstance(parameter_value.value, int):
      proto.value.number_value = parameter_value.value
    elif isinstance(parameter_value.value, bool):
      proto.value.bool_value = parameter_value.value
    elif isinstance(parameter_value.value, float):
      proto.value.number_value = parameter_value.value
    elif isinstance(parameter_value.value, str):
      proto.value.string_value = parameter_value.value

    return proto


class MeasurementConverter:
  """Converter for vz.Measurement."""

  @classmethod
  def from_proto(cls, proto: study_pb2.Measurement) -> trial.Measurement:
    """Creates a valid instance from proto.

    Args:
      proto: Measurement proto.

    Returns:
      A valid instance of Measurement object. Metrics with invalid values
      are automatically filtered out.
    """

    metrics = dict()

    for metric in proto.metrics:
      if (
          metric.metric_id in metrics
          and metrics[metric.metric_id].value != metric.value
      ):
        logging.log_first_n(
            logging.ERROR,
            (
                'Duplicate metric of name "%s".'
                'The newly found value %s will be used and '
                'the previously found value %s will be discarded.'
                'This always happens if the proto has an empty-named metric.'
            ),
            5,
            metric.metric_id,
            metric.value,
            metrics[metric.metric_id].value,
        )
      try:
        metrics[metric.metric_id] = trial.Metric(value=metric.value)
      except ValueError:
        pass
    return trial.Measurement(
        metrics=metrics,
        elapsed_secs=proto.elapsed_duration.seconds,
        steps=proto.step_count,
    )

  @classmethod
  def to_proto(cls, measurement: trial.Measurement) -> study_pb2.Measurement:
    """Converts to Measurement proto."""
    proto = study_pb2.Measurement()
    for name, metric in measurement.metrics.items():
      proto.metrics.add(metric_id=name, value=metric.value)

    proto.step_count = measurement.steps
    int_seconds = int(measurement.elapsed_secs)
    proto.elapsed_duration.seconds = int_seconds
    proto.elapsed_duration.nanos = int(
        1e9 * (measurement.elapsed_secs - int_seconds)
    )
    return proto


class MetricInformationConverter:
  """A converter to/from study_pb2.StudySpec.MetricInformation."""

  @classmethod
  def from_proto(
      cls, proto: study_pb2.StudySpec.MetricSpec
  ) -> base_study_config.MetricInformation:
    """Converts a MetricInformation proto to a MetricInformation object."""
    if proto.goal not in list(base_study_config.ObjectiveMetricGoal):
      raise ValueError('Unknown MetricInformation.goal: {}'.format(proto.goal))

    safety_threshold = None
    desired_min_safe_trials_fraction = None

    if proto.HasField('safety_config'):
      safety_threshold = proto.safety_config.safety_threshold
    if proto.safety_config.HasField('desired_min_safe_trials_fraction'):
      desired_min_safe_trials_fraction = (
          proto.safety_config.desired_min_safe_trials_fraction
      )

    return base_study_config.MetricInformation(
        name=proto.metric_id,
        goal=proto.goal,
        safety_threshold=safety_threshold,
        desired_min_safe_trials_fraction=desired_min_safe_trials_fraction,
        min_value=None,
        max_value=None,
    )

  @classmethod
  def to_proto(
      cls, obj: base_study_config.MetricInformation
  ) -> study_pb2.StudySpec.MetricSpec:
    """Returns this object as a proto."""

    proto = study_pb2.StudySpec.MetricSpec(
        metric_id=obj.name, goal=obj.goal.value
    )

    if obj.type == base_study_config.MetricType.SAFETY:
      proto.safety_config.safety_threshold = obj.safety_threshold
      if obj.desired_min_safe_trials_fraction is not None:
        proto.safety_config.desired_min_safe_trials_fraction = (
            obj.desired_min_safe_trials_fraction
        )
    return proto


class SearchSpaceConverter:
  """A wrapper for study_pb2.StudySpec."""

  @classmethod
  def from_proto(
      cls, proto: study_pb2.StudySpec
  ) -> parameter_config.SearchSpace:
    """Extracts a SearchSpace object from a StudyConfig proto."""
    space = parameter_config.SearchSpace()
    for pc in proto.parameters:
      space.add(ParameterConfigConverter.from_proto(pc))
    return space

  @classmethod
  def parameter_protos(
      cls, obj: parameter_config.SearchSpace
  ) -> List[study_pb2.StudySpec.ParameterSpec]:
    """Returns the search space as a List of ParameterConfig protos."""
    return [ParameterConfigConverter.to_proto(pc) for pc in obj.parameters]


class MetricsConfigConverter:
  """A wrapper for study_pb2.StudySpec.MetricSpec's."""

  @classmethod
  def from_protos(
      cls, protos: Iterable[study_pb2.StudySpec.MetricSpec]
  ) -> base_study_config.MetricsConfig:
    return base_study_config.MetricsConfig(
        [MetricInformationConverter.from_proto(m) for m in protos]
    )

  @classmethod
  def to_protos(
      cls, obj: base_study_config.MetricsConfig
  ) -> List[study_pb2.StudySpec.MetricSpec]:
    return [MetricInformationConverter.to_proto(metric) for metric in obj]


def _to_pyvizier_trial_status(
    proto_state: study_pb2.Trial.State,
) -> trial.TrialStatus:
  """from_proto conversion for Trial statuses."""
  if proto_state == study_pb2.Trial.State.REQUESTED:
    return trial.TrialStatus.REQUESTED
  elif proto_state == study_pb2.Trial.State.ACTIVE:
    return trial.TrialStatus.ACTIVE
  if proto_state == study_pb2.Trial.State.STOPPING:
    return trial.TrialStatus.STOPPING
  if proto_state == study_pb2.Trial.State.SUCCEEDED:
    return trial.TrialStatus.COMPLETED
  elif proto_state == study_pb2.Trial.State.INFEASIBLE:
    return trial.TrialStatus.COMPLETED
  else:
    return trial.TrialStatus.UNKNOWN


def _from_pyvizier_trial_status(
    status: trial.TrialStatus, infeasible: bool
) -> study_pb2.Trial.State:
  """to_proto conversion for Trial states."""
  if status == trial.TrialStatus.REQUESTED:
    return study_pb2.Trial.State.REQUESTED
  elif status == trial.TrialStatus.ACTIVE:
    return study_pb2.Trial.State.ACTIVE
  elif status == trial.TrialStatus.STOPPING:
    return study_pb2.Trial.State.STOPPING
  elif status == trial.TrialStatus.COMPLETED:
    if infeasible:
      return study_pb2.Trial.State.INFEASIBLE
    else:
      return study_pb2.Trial.State.SUCCEEDED
  else:
    return study_pb2.Trial.State.STATE_UNSPECIFIED


class TrialConverter:
  """Converter for vz.Trial."""

  @classmethod
  def from_proto(cls, proto: study_pb2.Trial) -> trial.Trial:
    """Converts from Trial proto to object.

    Args:
      proto: Trial proto.

    Returns:
      A Trial object.
    """
    parameters = {}
    for parameter in proto.parameters:
      value = ParameterValueConverter.from_proto(parameter)
      if value is not None:
        if parameter.parameter_id in parameters:
          raise ValueError(
              'Invalid trial proto contains duplicate parameter {}: {}'.format(
                  parameter.parameter_id, proto
              )
          )
        parameters[parameter.parameter_id] = value
      else:
        logging.warning(
            'A parameter without a value will be dropped: %s', parameter
        )

    final_measurement = None
    if proto.HasField('final_measurement'):
      final_measurement = MeasurementConverter.from_proto(
          proto.final_measurement
      )

    completion_time = None
    infeasibility_reason = None
    if proto.state == study_pb2.Trial.State.SUCCEEDED:
      if proto.HasField('end_time'):
        completion_ts = proto.end_time.seconds + 1e-9 * proto.end_time.nanos
        completion_time = datetime.datetime.fromtimestamp(completion_ts)
    elif proto.state == study_pb2.Trial.State.INFEASIBLE:
      infeasibility_reason = proto.infeasible_reason

    metadata = common.Metadata()
    for kv in proto.metadata:
      metadata.abs_ns(common.Namespace.decode(kv.ns))[kv.key] = (
          kv.proto if kv.HasField('proto') else kv.value
      )

    measurements = []
    for measure in proto.measurements:
      measurements.append(MeasurementConverter.from_proto(measure))

    creation_time = None
    if proto.HasField('start_time'):
      creation_ts = proto.start_time.seconds + 1e-9 * proto.start_time.nanos
      creation_time = datetime.datetime.fromtimestamp(creation_ts)
    return trial.Trial(
        id=int(proto.id),
        description=proto.name,
        assigned_worker=proto.client_id or None,
        is_requested=proto.state == proto.REQUESTED,
        stopping_reason=(
            'stopping reason not supported yet'
            if proto.state == proto.STOPPING
            else None
        ),
        parameters=parameters,
        creation_time=creation_time,
        completion_time=completion_time,
        infeasibility_reason=infeasibility_reason,
        final_measurement=final_measurement,
        measurements=measurements,
        metadata=metadata,
    )  # pytype: disable=wrong-arg-types

  @classmethod
  def from_protos(cls, protos: Iterable[study_pb2.Trial]) -> List[trial.Trial]:
    """Convenience wrapper for from_proto."""
    return [TrialConverter.from_proto(proto) for proto in protos]

  @classmethod
  def to_protos(cls, pytrials: Iterable[trial.Trial]) -> List[study_pb2.Trial]:
    return [TrialConverter.to_proto(pytrial) for pytrial in pytrials]

  @classmethod
  def to_proto(cls, pytrial: trial.Trial) -> study_pb2.Trial:
    """Converts a pyvizier Trial to a Trial proto."""
    proto = study_pb2.Trial()
    if pytrial.description is not None:
      proto.name = pytrial.description
    proto.id = str(pytrial.id)
    proto.state = _from_pyvizier_trial_status(
        pytrial.status, pytrial.infeasible
    )
    proto.client_id = pytrial.assigned_worker or ''

    for name, value in pytrial.parameters.items():
      proto.parameters.append(ParameterValueConverter.to_proto(value, name))

    # pytrial always adds an empty metric. Ideally, we should remove it if the
    # metric does not exist in the study config.
    if pytrial.final_measurement is not None:
      proto.final_measurement.CopyFrom(
          MeasurementConverter.to_proto(pytrial.final_measurement)
      )

    for measurement in pytrial.measurements:
      proto.measurements.append(MeasurementConverter.to_proto(measurement))

    if pytrial.creation_time is not None:
      creation_secs = datetime.datetime.timestamp(pytrial.creation_time)
      proto.start_time.seconds = int(creation_secs)
      proto.start_time.nanos = int(1e9 * (creation_secs - int(creation_secs)))
    if pytrial.completion_time is not None:
      completion_secs = datetime.datetime.timestamp(pytrial.completion_time)
      proto.end_time.seconds = int(completion_secs)
      proto.end_time.nanos = int(1e9 * (completion_secs - int(completion_secs)))
    if pytrial.infeasibility_reason is not None:
      proto.infeasible_reason = pytrial.infeasibility_reason
    if pytrial.metadata is not None:
      for ns in pytrial.metadata.namespaces():
        ns_string = ns.encode()
        ns_layer = pytrial.metadata.abs_ns(ns)
        for key, value in ns_layer.items():
          metadata_util.assign(proto, key=key, ns=ns_string, value=value)
    return proto


class TrialSuggestionConverter:
  """Converts vz.TrialSuggestion <--> Pythia TrialSuggestion proto."""

  @classmethod
  def from_proto(
      cls, proto: pythia_service_pb2.TrialSuggestion
  ) -> trial.TrialSuggestion:
    """Converts from TrialSuggestion proto to PyVizier TrialSuggestion."""
    parameters = {}
    for parameter in proto.parameters:
      value = ParameterValueConverter.from_proto(parameter)
      if value is None:
        raise RuntimeError('Parameter %s exists without a value.' % parameter)
      if parameter.parameter_id in parameters:
        raise ValueError(
            'Invalid trial proto contains duplicate parameter {}: {}'.format(
                parameter.parameter_id, proto
            )
        )
      parameters[parameter.parameter_id] = value

    metadata = common.Metadata()
    for kv in proto.metadata:
      metadata.abs_ns(common.Namespace.decode(kv.ns))[kv.key] = (
          kv.proto if kv.HasField('proto') else kv.value
      )

    return trial.TrialSuggestion(parameters=parameters, metadata=metadata)

  @classmethod
  def from_protos(
      cls, protos: Iterable[pythia_service_pb2.TrialSuggestion]
  ) -> List[trial.TrialSuggestion]:
    """Convenience wrapper for from_proto."""
    return [cls.from_proto(proto) for proto in protos]

  @classmethod
  def to_protos(
      cls, pytrials: Iterable[trial.TrialSuggestion]
  ) -> List[pythia_service_pb2.TrialSuggestion]:
    return [cls.to_proto(pytrial) for pytrial in pytrials]

  @classmethod
  def to_proto(
      cls, suggestion: trial.TrialSuggestion
  ) -> pythia_service_pb2.TrialSuggestion:
    """Converts a pyvizier TrialSuggestion to the corresponding proto."""
    proto = pythia_service_pb2.TrialSuggestion()

    for name, value in suggestion.parameters.items():
      proto.parameters.append(ParameterValueConverter.to_proto(value, name))

    proto.metadata.extend(
        metadata_util.make_key_value_list(suggestion.metadata)
    )
    return proto


class MetadataDeltaConverter:
  """Converts pyvizier.MetadataDelta <--> List of UnitMetadataUpdate protos."""

  @classmethod
  def to_protos(
      cls, delta: trial.MetadataDelta
  ) -> List[vizier_service_pb2.UnitMetadataUpdate]:
    """Converts pyvizier.MetadataDelta to a List of UnitMetadataUpdate protos."""
    unit_metadata_updates = metadata_util.study_metadata_to_update_list(
        delta.on_study
    )
    unit_metadata_updates.extend(
        metadata_util.trial_metadata_to_update_list(delta.on_trials)
    )
    return unit_metadata_updates

  @classmethod
  def from_protos(
      cls, protos: Iterable[vizier_service_pb2.UnitMetadataUpdate]
  ) -> trial.MetadataDelta:
    """Converts a list of UnitMetadataUpdate protos to pyvizier.MetadataDelta."""
    mdd = trial.MetadataDelta()
    for u_m_u in protos:
      key_value = u_m_u.metadatum
      namespace = common.Namespace.decode(key_value.ns)
      value = (
          key_value.proto if key_value.HasField('proto') else key_value.value
      )
      if u_m_u.HasField('trial_id'):
        mdd.on_trials[int(u_m_u.trial_id)].abs_ns(namespace)[
            key_value.key
        ] = value
      else:
        mdd.on_study.abs_ns(namespace)[key_value.key] = value
    return mdd


class ProblemStatementConverter:
  """Converts pyvizier.ProblemStatement <-> Pythia ProblemStatement proto."""

  @classmethod
  def to_proto(
      cls,
      problem_statement: base_study_config.ProblemStatement,
  ) -> pythia_service_pb2.ProblemStatement:
    """Converts PyVizier ProblemStatement to Proto version."""
    parameter_spec_protos = SearchSpaceConverter.parameter_protos(
        problem_statement.search_space
    )
    metric_information_protos = MetricsConfigConverter.to_protos(
        problem_statement.metric_information
    )
    keyvalue_protos = metadata_util.make_key_value_list(
        problem_statement.metadata
    )
    return pythia_service_pb2.ProblemStatement(
        search_space=parameter_spec_protos,
        metric_information=metric_information_protos,
        metadata=keyvalue_protos,
    )

  @classmethod
  def from_proto(
      cls, proto: pythia_service_pb2.ProblemStatement
  ) -> base_study_config.ProblemStatement:
    """Converts ProblemStatement Proto to PyVizier version."""
    study_spec = study_pb2.StudySpec(parameters=proto.search_space)
    search_space = SearchSpaceConverter.from_proto(study_spec)
    metric_information = MetricsConfigConverter.from_protos(
        proto.metric_information
    )
    metadata = metadata_util.from_key_value_list(proto.metadata)
    return base_study_config.ProblemStatement(
        search_space=search_space,
        metric_information=metric_information,
        metadata=metadata,
    )


class StudyDescriptorConverter:
  """Converts Pythia StudyDescriptorConverter <-> Pythia StudyDescriptor proto."""

  @classmethod
  def to_proto(
      cls,
      study_descriptor: study.StudyDescriptor,
  ) -> pythia_service_pb2.StudyDescriptor:
    return pythia_service_pb2.StudyDescriptor(
        config=ProblemStatementConverter.to_proto(study_descriptor.config),
        guid=study_descriptor.guid,
        max_trial_id=study_descriptor.max_trial_id,
    )

  @classmethod
  def from_proto(
      cls, proto: pythia_service_pb2.StudyDescriptor
  ) -> study.StudyDescriptor:
    return study.StudyDescriptor(
        config=ProblemStatementConverter.from_proto(proto.config),
        guid=proto.guid,
        max_trial_id=proto.max_trial_id,
    )


class SuggestConverter:
  """Converts a SuggestRequest class <--> a SuggestRequest proto."""

  @classmethod
  def to_request_proto(
      cls,
      request: policy.SuggestRequest,
  ) -> pythia_service_pb2.SuggestRequest:
    """Conversion from PyVizier to proto."""
    study_descriptor_proto = StudyDescriptorConverter.to_proto(
        request._study_descriptor  # pylint:disable=protected-access
    )
    return pythia_service_pb2.SuggestRequest(
        study_descriptor=study_descriptor_proto,
        count=request.count,
        checkpoint_dir=request.checkpoint_dir,
    )

  @classmethod
  def from_request_proto(
      cls,
      proto: pythia_service_pb2.SuggestRequest,
  ) -> policy.SuggestRequest:
    """Conversion from proto to PyVizier."""
    study_descriptor = StudyDescriptorConverter.from_proto(
        proto.study_descriptor
    )
    return policy.SuggestRequest(
        study_descriptor=study_descriptor,
        count=proto.count,
        checkpoint_dir=proto.checkpoint_dir,
    )

  @classmethod
  def to_decision_proto(
      cls,
      decision: policy.SuggestDecision,
  ) -> pythia_service_pb2.SuggestDecision:
    """Conversion from PyVizier to proto."""
    trial_suggestion_protos = TrialSuggestionConverter.to_protos(
        decision.suggestions
    )
    metadelta_protos = MetadataDeltaConverter.to_protos(decision.metadata)
    return pythia_service_pb2.SuggestDecision(
        suggestions=trial_suggestion_protos, metadata=metadelta_protos
    )

  @classmethod
  def from_decision_proto(
      cls,
      proto: pythia_service_pb2.SuggestDecision,
  ) -> policy.SuggestDecision:
    """Conversion from proto to PyVizier."""
    suggestions = TrialSuggestionConverter.from_protos(proto.suggestions)
    metadata = MetadataDeltaConverter.from_protos(proto.metadata)
    return policy.SuggestDecision(suggestions=suggestions, metadata=metadata)


class EarlyStopConverter:
  """Converts Pythia EarlyStopping <-> Pythia EarlyStopping protos."""

  @classmethod
  def to_request_proto(
      cls,
      request: policy.EarlyStopRequest,
  ) -> pythia_service_pb2.EarlyStopRequest:
    """Conversion from PyVizier to proto."""
    return pythia_service_pb2.EarlyStopRequest(
        study_descriptor=StudyDescriptorConverter.to_proto(
            request._study_descriptor  # pylint:disable=protected-access
        ),
        trial_ids=request.trial_ids,
        checkpoint_dir=request.checkpoint_dir,
    )

  @classmethod
  def from_request_proto(
      cls,
      proto: pythia_service_pb2.EarlyStopRequest,
  ) -> policy.EarlyStopRequest:
    """Conversion from proto to PyVizier."""
    study_descriptor = StudyDescriptorConverter.from_proto(
        proto.study_descriptor
    )
    return policy.EarlyStopRequest(
        study_descriptor=study_descriptor,
        trial_ids=proto.trial_ids,
        checkpoint_dir=proto.checkpoint_dir,
    )

  @classmethod
  def to_decisions_proto(
      cls,
      decisions: policy.EarlyStopDecisions,
  ) -> pythia_service_pb2.EarlyStopDecisions:
    """Conversion from PyVizier to proto."""
    decision_protos = []
    for decision in decisions.decisions:
      predicted_final_measurement_proto = study_pb2.Measurement()
      if decision.predicted_final_measurement:
        predicted_final_measurement_proto = MeasurementConverter.to_proto(
            decision.predicted_final_measurement
        )
      decision_proto = pythia_service_pb2.EarlyStopDecision(
          id=decision.id,
          reason=decision.reason,
          should_stop=decision.should_stop,
          predicted_final_measurement=predicted_final_measurement_proto,
      )
      decision_protos.append(decision_proto)
    key_value_protos = MetadataDeltaConverter.to_protos(decisions.metadata)
    return pythia_service_pb2.EarlyStopDecisions(
        decisions=decision_protos, metadata=key_value_protos
    )

  @classmethod
  def from_decisions_proto(
      cls,
      proto: pythia_service_pb2.EarlyStopDecisions,
  ) -> policy.EarlyStopDecisions:
    """Conversion from proto to PyVizier."""
    decisions = []
    for decision_proto in proto.decisions:
      decision = policy.EarlyStopDecision(
          id=decision_proto.id,
          reason=decision_proto.reason,
          should_stop=decision_proto.should_stop,
          predicted_final_measurement=MeasurementConverter.from_proto(
              decision_proto.predicted_final_measurement
          ),
      )
      decisions.append(decision)
    metadata = MetadataDeltaConverter.from_protos(proto.metadata)
    return policy.EarlyStopDecisions(decisions=decisions, metadata=metadata)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Lightweight experimenters."""

from vizier._src.benchmarks.experimenters import experimenter_factory
from vizier._src.benchmarks.experimenters.combo_experimenter import CentroidExperimenter
from vizier._src.benchmarks.experimenters.combo_experimenter import ContaminationExperimenter
from vizier._src.benchmarks.experimenters.combo_experimenter import IsingExperimenter
from vizier._src.benchmarks.experimenters.combo_experimenter import MAXSATExperimenter
from vizier._src.benchmarks.experimenters.combo_experimenter import PestControlExperimenter
from vizier._src.benchmarks.experimenters.discretizing_experimenter import DiscretizingExperimenter
from vizier._src.benchmarks.experimenters.experimenter import Experimenter
from vizier._src.benchmarks.experimenters.experimenter_factory import BBOBExperimenterFactory
from vizier._src.benchmarks.experimenters.experimenter_factory import ExperimenterFactory
from vizier._src.benchmarks.experimenters.experimenter_factory import SerializableExperimenterFactory
from vizier._src.benchmarks.experimenters.experimenter_factory import SingleObjectiveExperimenterFactory
from vizier._src.benchmarks.experimenters.infeasible_experimenter import HashingInfeasibleExperimenter
from vizier._src.benchmarks.experimenters.infeasible_experimenter import ParamRegionInfeasibleExperimenter
from vizier._src.benchmarks.experimenters.l1_categorical_experimenter import L1CategorialExperimenter
from vizier._src.benchmarks.experimenters.multiobjective_experimenter import MultiObjectiveExperimenter
from vizier._src.benchmarks.experimenters.noisy_experimenter import NoisyExperimenter
from vizier._src.benchmarks.experimenters.normalizing_experimenter import NormalizingExperimenter
from vizier._src.benchmarks.experimenters.numpy_experimenter import NumpyExperimenter
from vizier._src.benchmarks.experimenters.permuting_experimenter import PermutingExperimenter
from vizier._src.benchmarks.experimenters.shifting_experimenter import ShiftingExperimenter
from vizier._src.benchmarks.experimenters.sign_flip_experimenter import SignFlipExperimenter
from vizier._src.benchmarks.experimenters.sparse_experimenter import SparseExperimenter
from vizier._src.benchmarks.experimenters.surrogate_experimenter import PredictorExperimenter
from vizier._src.benchmarks.experimenters.switch_experimenter import SwitchExperimenter
from vizier._src.benchmarks.experimenters.synthetic import bbob
from vizier._src.benchmarks.experimenters.synthetic.branin import Branin2DExperimenter
from vizier._src.benchmarks.experimenters.synthetic.hartmann import Hartmann3DExperimenter
from vizier._src.benchmarks.experimenters.synthetic.hartmann import Hartmann6DExperimenter
from vizier._src.benchmarks.experimenters.synthetic.hartmann import HartmannExperimenter
from vizier._src.benchmarks.experimenters.synthetic.simplekd import SimpleKDExperimenter



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Wrapper classes for Trial protos and other messages in them."""

import collections
from collections import abc
import copy
import dataclasses
import datetime
import enum
from typing import Any, Dict, List, Mapping, MutableMapping, Optional, Union, FrozenSet

from absl import logging
import attr
import numpy as np

from vizier._src.pyvizier.shared import common

ParameterValueTypes = Union[str, int, float, bool]

# TODO: These constants should be deleted.
TRUE_VALUE = 'True'
FALSE_VALUE = 'False'


class ParameterType(enum.Enum):
  """Valid Values for ParameterConfig.type."""
  DOUBLE = 'DOUBLE'
  INTEGER = 'INTEGER'
  CATEGORICAL = 'CATEGORICAL'
  DISCRETE = 'DISCRETE'
  CUSTOM = 'CUSTOM'

  def is_numeric(self) -> bool:
    return self in [self.DOUBLE, self.INTEGER, self.DISCRETE]

  def is_continuous(self) -> bool:
    return self == self.DOUBLE

  def _raise_type_error(self, value: ParameterValueTypes) -> None:
    raise TypeError(f'Type {self} is not compatible with value: {value}')

  def assert_correct_type(self, value: ParameterValueTypes) -> None:
    if self.is_numeric() and float(value) != value:
      self._raise_type_error(value)

    # TODO: Accepting boolean into categorical is unintuitive.
    elif (self
          == ParameterType.CATEGORICAL) and (not isinstance(value,
                                                            (str, bool))):
      self._raise_type_error(value)

    if self == self.INTEGER and int(value) != value:
      self._raise_type_error(value)


# TODO: Trial class should not depend on these.
class ExternalType(enum.Enum):
  """Valid Values for ParameterConfig.external_type."""
  INTERNAL = 'INTERNAL'
  BOOLEAN = 'BOOLEAN'
  INTEGER = 'INTEGER'
  FLOAT = 'FLOAT'


# Values should NEVER be removed from the enums below, only added.
class TrialStatus(enum.Enum):
  """Values for Trial.Status."""
  UNKNOWN = 'UNKNOWN'
  REQUESTED = 'REQUESTED'
  ACTIVE = 'ACTIVE'
  COMPLETED = 'COMPLETED'
  STOPPING = 'STOPPING'


@attr.s(frozen=True, init=True, slots=True, kw_only=False)
class Metric:
  """Enhanced immutable wrapper for vizier_pb2.Metric proto.

  It has an optional field "std" for internal usage. This field gets lost
  when the object is converted to proto.
  """

  def _std_not_negative(self, _, stddev: Optional[float]) -> bool:
    if (stddev is not None) and (not stddev >= 0):
      raise ValueError(
          'Standard deviation must be a non-negative finite number.')

  value: float = attr.ib(
      converter=float,
      init=True,
      validator=[attr.validators.instance_of(float)],
      kw_only=False)
  std: Optional[float] = attr.ib(
      converter=lambda x: float(x) if x is not None else None,
      validator=[
          attr.validators.optional(attr.validators.instance_of(float)),
          _std_not_negative
      ],
      init=True,
      default=None,
      kw_only=True)


# Use when you want to preserve the shapes or reduce if-else statements.
# e.g. `metrics.get('metric_name', NaNMetric).value` to get NaN or the actual
# value.
NaNMetric = Metric(value=np.nan)


# TODO: This class should be deleted in the future.
@attr.s(auto_attribs=True, frozen=True, init=True, slots=True, repr=False)
class ParameterValue:
  """Immutable wrapper for vizier_pb2.Parameter.value, which is a oneof field.

  Has accessors (properties) that cast the value into the type according
  to StudyConfiguration class behavior. In particular, 'true' and FALSE_VALUE
  are
  treated as special strings that are cast to a numeric value of 1 and 0,
  respectively, and boolean value of True and False, repectively.
  """

  value: ParameterValueTypes = attr.ib(
      init=True,
      validator=[
          attr.validators.instance_of((str, int, float, bool)),
      ])

  def cast_as_internal(self,
                       internal_type: ParameterType) -> ParameterValueTypes:
    """Cast to the internal type."""
    internal_type.assert_correct_type(self.value)

    if internal_type in (ParameterType.DOUBLE, ParameterType.DISCRETE):
      return self.as_float
    elif internal_type == ParameterType.INTEGER:
      return self.as_int
    elif internal_type == ParameterType.CATEGORICAL:
      return self.as_str
    else:
      raise RuntimeError(f'Unknown type {internal_type}')

  def cast(
      self,
      external_type: ExternalType,
  ) -> ParameterValueTypes:
    """Returns ParameterValue cast to external_type.

    Args:
      external_type:

    Returns:
      self.value if external_type is INTERNAL.
      self.as_bool if external_type is BOOLEAN.
      self.as_int if external_type is INTEGER.
      self.as_float if external_type is FLOAT.

    Raises:
      ValueError: If external_type is not valid.
    """
    if external_type == ExternalType.INTERNAL:
      return self.value
    elif external_type == ExternalType.BOOLEAN:
      return self.as_bool
    elif external_type == ExternalType.INTEGER:
      return self.as_int
    elif external_type == ExternalType.FLOAT:
      return self.as_float
    else:
      raise ValueError(
          'Unknown external type enum value: {}.'.format(external_type))

  @property
  def as_float(self) -> Optional[float]:
    """Returns the value cast to float."""
    if self.value == TRUE_VALUE:
      return 1.0
    elif self.value == FALSE_VALUE:
      return 0.0
    try:
      # Note str -> float conversion exists for benchmark use.
      return float(self.value)
    except ValueError:
      return None

  @property
  def as_int(self) -> Optional[int]:
    """Returns the value cast to int."""
    if self.value == TRUE_VALUE:
      return 1
    elif self.value == FALSE_VALUE:
      return 0
    try:
      # Note str -> int conversion exists for benchmark use.
      return int(self.value)
    except ValueError:
      return None

  @property
  def as_str(self) -> Optional[str]:
    """Returns str-typed value or 'True'/'False' if value is bool."""
    if isinstance(self.value, bool):
      if self.value:
        return TRUE_VALUE
      else:
        return FALSE_VALUE
    elif isinstance(self.value, str):
      return self.value
    # Note str conversion exists for benchmark use. Use __str__ instead.
    return str(self.value)

  @property
  def as_bool(self) -> Optional[bool]:
    """Returns the value as bool following StudyConfiguration's behavior.

    Returns: True if value is TRUE_VALUE or 1. False if value is
      FALSE_VALUE or 0. For all other cases, returns None.
      For string type, this behavior is consistent with how
      StudyConfiguration.AddBooleanParameter's. For other types, this
      guarantees that self.value == self.as_bool
    """
    if isinstance(self.value, str):
      if self.value == TRUE_VALUE:
        return True
      elif self.value == FALSE_VALUE:
        return False
    else:
      if self.value == 1.0:
        return True
      elif self.value == 0.0:
        return False
    return None

  def __str__(self) -> str:
    return str(self.value)

  def __repr__(self) -> str:
    return repr(self.value)


class _MetricDict(collections.UserDict, Mapping[str, Metric]):
  """Dictionary of string to metrics."""

  def get_value(self, key: str, default: float | None) -> float | None:
    if key in self.data:
      return self.data[key].value
    else:
      return default

  def __setitem__(self, key: str, value: Union[float, Metric]):
    if isinstance(value, Metric):
      self.data.__setitem__(key, value)
    else:
      self.data.__setitem__(key, Metric(value=value))

  def as_float_dict(self) -> dict[str, float]:
    return {k: m.value for k, m in self.data.items()}


@attr.s(auto_attribs=True, frozen=False, init=True, slots=True)
class Measurement:
  """A collection of metrics with a timestamp & checkpoint.

  metrics: Named, floating-point metrics.  A typical example would be the
    accuracy of a machine-learning model.  Typically, all the metrics mentioned
    in the MetricInformation class would be listed here; other metrics may be
    listed but would not normally be used by Vizier.

  elapsed_secs: (optional) The length of time it took to evaluate the Trial to
    reach this Measurement, in seconds.  This may be used by some Vizier
    algorithms.

  steps: (optional)  A positive integer roughly proportional to the amount of
    work spent.  When training a ML system, often this is a count of training
    steps/epochs.  When supplied, $steps should be consistent with the order of
    Measurements, but they need not be consecutive values.

  checkpoint_path: (optional) A slash-separated pathname.  Typically used when
    training a ML model; it would normally be a pathname for the checkpoint that
    produced the Measurement.  Implementations may limit the length of this
    string, but at least 2048 bytes will be allowed.
  """

  def _value_is_finite(self, _, value):
    if not (np.isfinite(value) and value >= 0):
      raise ValueError('Must be finite and non-negative.')

  # Should be used as a regular Dict.
  metrics: _MetricDict = attr.ib(
      init=True,
      converter=lambda d: _MetricDict(**d),
      default=_MetricDict(),
      validator=attr.validators.instance_of(_MetricDict),
      on_setattr=[attr.setters.convert, attr.setters.validate])

  elapsed_secs: float = attr.ib(
      converter=float,
      init=True,
      default=0,
      validator=[attr.validators.instance_of(float), _value_is_finite],
      on_setattr=[attr.setters.convert, attr.setters.validate],
      kw_only=True)

  # TODO: Change type annotation to int.
  steps: float = attr.ib(
      converter=int,
      init=True,
      default=0,
      validator=[attr.validators.instance_of(int), _value_is_finite],
      on_setattr=[attr.setters.convert, attr.setters.validate],
      kw_only=True)

  checkpoint_path: str = attr.ib(
      init=True,
      default='',
      validator=[attr.validators.instance_of(str)],
      kw_only=True,
  )


def _to_local_time(
    dt: Optional[datetime.datetime]) -> Optional[datetime.datetime]:
  """Converter for initializing timestamps in Trial class."""
  return dt.astimezone() if dt else None


# TODO: This class should have group() method that
# groups list parameters under the same key.
@attr.define(init=False, frozen=True, eq=True)
class ParameterDict(abc.MutableMapping):
  """Parameter dictionary.

  Maps the parameter names to their values. Works like a regular
  dict[str, ParameterValue] for the most part, except one can directly assign
  values of type `ParameterValueType`. So,
    ParameterDict(a=3) and
    ParameterDict(a=ParameterValue(3)) are equivalent.


  To access the raw value directly, use get_value() or as_dict():
    d.get_value('a') == d.get('a').value
    d.as_dict()['a'] == d.get_value('a')
  """

  _items: MutableMapping[str, ParameterValue] = attr.field(
      init=False, factory=dict)

  def as_dict(self) -> Dict[str, ParameterValueTypes]:
    """Returns the dict of parameter names to raw values."""
    return {k: self.get_value(k) for k in self._items}

  def __init__(self, iterable: Any = tuple(), **kwargs):
    self.__attrs_init__()
    self.update(iterable, **kwargs)

  def __setitem__(self, key: str, value: Union[ParameterValue,
                                               ParameterValueTypes]):
    if isinstance(value, ParameterValue):
      self._items[key] = value
    else:
      self._items[key] = ParameterValue(value)

  def __delitem__(self, key: str):
    del self._items[key]

  def __getitem__(self, key: str) -> ParameterValue:
    return self._items[key]

  def __len__(self) -> int:
    return len(self._items)

  def __iter__(self):
    return iter(self._items)

  def get_value(
      self,
      key: str,
      default: Optional[ParameterValueTypes] = None
  ) -> Optional[ParameterValueTypes]:
    """Returns the raw value of the given parameter name."""
    pv = self.get(key, default)
    if isinstance(pv, ParameterValue):
      return pv.value
    else:
      return pv


@attr.define(auto_attribs=True, frozen=False, init=True, slots=True)
class TrialSuggestion:
  """Freshly suggested trial.

  Suggestion can be converted to Trial object which has more functionalities.
  """

  parameters: ParameterDict = attr.field(
      init=True,
      factory=ParameterDict,
      converter=ParameterDict,
      validator=attr.validators.instance_of(ParameterDict))  # pytype: disable=wrong-arg-types

  metadata: common.Metadata = attr.field(
      init=True,
      kw_only=True,
      factory=common.Metadata,
      validator=attr.validators.instance_of(common.Metadata))

  def to_trial(self, uid: int = 0) -> 'Trial':
    """Assign an id and make it a Trial object.

    Usually SuggetedTrial objects are shorted-lived and not exposed to end
    users. This method is for non-service usage of trial suggestions in
    benchmarks, tests, colabs, etc.

    Args:
      uid: Trial id.

    Returns:
      Trial object.
    """
    return Trial(id=uid, parameters=self.parameters, metadata=self.metadata)


@attr.define(auto_attribs=True, frozen=False, init=True, slots=True)
class Trial(TrialSuggestion):
  """A Vizier Trial."""
  id: int = attr.ib(
      init=True,
      kw_only=True,
      default=0,
      validator=attr.validators.instance_of(int),
  )

  is_requested: bool = attr.ib(
      init=True,
      kw_only=True,
      default=False,
      validator=attr.validators.instance_of(bool))

  assigned_worker: Optional[str] = attr.ib(
      init=True,
      kw_only=True,
      default=None,
      validator=attr.validators.optional(attr.validators.instance_of(str)),
  )

  stopping_reason: Optional[str] = attr.ib(
      init=True,
      kw_only=True,
      default=None,
      validator=attr.validators.optional(attr.validators.instance_of(str)),
  )

  _infeasibility_reason: Optional[str] = attr.ib(
      init=True,
      kw_only=True,
      default=None,
      validator=attr.validators.optional(attr.validators.instance_of(str)),
  )

  description: Optional[str] = attr.ib(
      init=True,
      kw_only=True,
      default=None,
      validator=attr.validators.optional(attr.validators.instance_of(str)),
  )

  related_links: Dict[str, str] = attr.ib(
      init=True,
      kw_only=True,
      factory=dict,
      validator=attr.validators.deep_mapping(
          key_validator=attr.validators.instance_of(str),
          value_validator=attr.validators.instance_of(str),
          mapping_validator=attr.validators.instance_of(dict)),
  )  # pytype: disable=wrong-arg-types

  final_measurement: Optional[Measurement] = attr.ib(
      init=True,
      kw_only=True,
      default=None,
      validator=attr.validators.optional(
          attr.validators.instance_of(Measurement)),
  )

  measurements: List[Measurement] = attr.ib(
      init=True,
      kw_only=True,
      factory=list,
      validator=attr.validators.deep_iterable(
          member_validator=attr.validators.instance_of(Measurement),
          iterable_validator=attr.validators.instance_of(list)),
  )

  creation_time: Optional[datetime.datetime] = attr.ib(
      init=True,
      factory=datetime.datetime.now,
      converter=_to_local_time,
      kw_only=True,
      repr=lambda v: v.strftime('%x %X') if v is not None else 'None',
      validator=attr.validators.optional(
          attr.validators.instance_of(datetime.datetime)
      ),
  )

  completion_time: Optional[datetime.datetime] = attr.ib(
      init=True,
      kw_only=True,
      default=None,
      eq=False,
      repr=lambda v: v.strftime('%x %X') if v is not None else 'None',
      converter=_to_local_time,
      validator=attr.validators.optional(
          attr.validators.instance_of(datetime.datetime)
      ),
  )

  def __attrs_post_init__(self):
    if self.completion_time is None and (
        self.final_measurement is not None or self.infeasibility_reason
    ):
      self.completion_time = self.creation_time

  @property
  def duration(self) -> Optional[datetime.timedelta]:
    """Returns the duration of this Trial if it is completed, or None."""
    if self.completion_time:
      return self.completion_time - self.creation_time
    else:
      return None

  @property
  def status(self) -> TrialStatus:
    """Status.

    COMPLETED: Trial has final measurement or is declared infeasible.
    ACTIVE: Trial is being evaluated.
    STOPPING: Trial is being evaluated, but was decided to be not worth further
      evaluating.
    REQUESTED: Trial is queued for future suggestions.
    """
    if self.final_measurement is not None or self.infeasible:
      return TrialStatus.COMPLETED
    elif self.stopping_reason is not None:
      return TrialStatus.STOPPING
    elif self.is_requested:
      return TrialStatus.REQUESTED
    else:
      return TrialStatus.ACTIVE

  @property
  def is_completed(self) -> bool:
    """Returns True if this Trial is completed."""
    if self.status == TrialStatus.COMPLETED:
      if self.completion_time is None:
        logging.warning('Invalid Trial state: status is COMPLETED, but a '
                        ' completion_time was not set')
      return True
    elif self.completion_time is not None:
      if self.status is None:
        logging.warning('Invalid Trial state: status is not set to COMPLETED, '
                        'but a completion_time is set')
      return True
    return False

  @property
  def infeasible(self) -> bool:
    """Returns True if this Trial is infeasible."""
    return self._infeasibility_reason is not None

  @property
  def infeasibility_reason(self) -> Optional[str]:
    """Returns this Trial's infeasibility reason, if set."""
    return self._infeasibility_reason

  def complete(self,
               measurement: Measurement,
               *,
               infeasibility_reason: Optional[str] = None,
               inplace: bool = True) -> 'Trial':
    """Completes the trial and returns it.

    Args:
      measurement: Measurement to complete the trial with.
      infeasibility_reason: If set, completes the trial as infeasible. If the
        trial was already infeasible and infeasibility_reason is not set, the
        trial remains infeasible.
      inplace: If True, Trial is modified in place. If False, which is the
        default, then the operation is performed and it returns a copy of the
        object.

    Returns:
      Completed Trial.
    """
    if inplace:
      # Use setattr. If we assign to self.final_measurement, then hyperref
      # mechanisms think this line is where `final_measurement` property
      # is defined, instead of where we declare attr.ib.
      self.__setattr__('final_measurement', copy.deepcopy(measurement))
      if infeasibility_reason is not None:
        self.__setattr__('_infeasibility_reason', infeasibility_reason)
      self.completion_time = _to_local_time(datetime.datetime.now())
      return self
    else:
      clone = copy.deepcopy(self)
      return clone.complete(
          measurement, inplace=True, infeasibility_reason=infeasibility_reason)

  @property
  def final_measurement_or_die(self) -> Measurement:
    if self.final_measurement is None:
      raise ValueError(f'Trial is missing final_measurement: {self}')
    return self.final_measurement


# Define aliases.
CompletedTrial = Trial
PendingTrial = Trial
CompletedTrialWithMeasurements = Trial
PendingTrialWithMeasurements = Trial


@attr.define(kw_only=True)
class TrialFilter:
  """Trial filter.

  All filters are by default 'AND' conditions.

  Attributes:
    ids: If set, requires the trial's id to be in the set.
    min_id: If set, requires the trial's id to be at least this number.
    max_id: If set, requires the trial's id to be at most this number.
    status: If set, requires the trial's status to be in the set.
  """
  ids: Optional[FrozenSet[int]] = attr.field(
      default=None,
      converter=lambda x: frozenset(x) if x is not None else None,
      validator=attr.validators.optional(
          attr.validators.deep_iterable(
              attr.validators.instance_of(int),
              attr.validators.instance_of(frozenset))))
  min_id: Optional[int] = attr.field(default=None)
  max_id: Optional[int] = attr.field(default=None)
  status: Optional[FrozenSet[TrialStatus]] = attr.field(
      default=None,
      converter=lambda x: frozenset(x) if x is not None else None,
      validator=attr.validators.optional(
          attr.validators.deep_iterable(
              attr.validators.instance_of(TrialStatus),
              attr.validators.instance_of(frozenset))))

  # TODO: Add "search_space" argument

  def __call__(self, trial: Trial) -> bool:
    if self.ids is not None:
      if trial.id not in self.ids:
        return False
    if self.min_id is not None:
      if trial.id < self.min_id:
        return False
    if self.max_id is not None:
      if trial.id > self.max_id:
        return False
    if self.status is not None:
      if trial.status not in self.status:
        return False
    return True


@dataclasses.dataclass(frozen=True)
class MetadataDelta:
  """Carries cumulative delta for a batch metadata update.

  Attributes:
    on_study: Updates to be made on study-level metadata.
    on_trials: Maps trial id to updates.
  """

  on_study: common.Metadata = dataclasses.field(default_factory=common.Metadata)

  on_trials: Dict[int, common.Metadata] = dataclasses.field(
      default_factory=lambda: collections.defaultdict(common.Metadata))

  def __bool__(self):
    """Returns True if this carries any Metadata items."""
    if self.on_study:
      return True
    for v in self.on_trials.values():
      if v:
        return True
    return False

  def on_trial(self, trial_id: int) -> common.Metadata:
    """Enables easy assignment to a single Trial."""
    return self.on_trials[trial_id]

  def assign(self,
             namespace: str,
             key: str,
             value: common.MetadataValue,
             *,
             trial: Optional[Trial] = None,
             trial_id: Optional[int] = None):
    """Assigns metadata.

    Args:
      namespace: Namespace of the metadata. See common.Metadata doc for more
        details.
      key:
      value:
      trial: If specified, `trial_id` must be None. It behaves the same as when
        `trial_id=trial.id`, and additionally, the metadata is added to `trial`.
      trial_id: If specified, `trial` must be None. If both `trial` and
        `trial_id` are None, then the key-value pair will be assigned to the
        study.

    Raises:
      ValueError:
    """
    if trial is None and trial_id is None:
      self.on_study.ns(namespace)[key] = value
    elif trial is not None and trial_id is not None:
      raise ValueError(
          'At most one of `trial` and `trial_id` can be specified.')
    elif trial is not None:
      self.on_trials[trial.id].ns(namespace)[key] = value
      trial.metadata.ns(namespace)[key] = value
    elif trial_id is not None:
      self.on_trials[trial_id].ns(namespace)[key] = value



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""High-level wrappers for stochastic process hyperparameter optimizers."""

# TODO: Add optimizers that are frequently used in the literature.

from typing import Optional
from absl import logging
import attr
import chex
import jax
from jax import numpy as jnp
import optax
from tensorflow_probability.substrates import jax as tfp
from vizier._src.jax import stochastic_process_model as sp
from vizier._src.jax.optimizers import core

tfb = tfp.bijectors

OptState = chex.ArrayTree


@attr.define
class OptaxTrain(core.Optimizer[core.Params]):
  """Wraps an Optax optimizer.

  It's recommended to use this optimizer with a loss function that normalizes by
  the number of observations. The unnormalized loss function for parameters `p`
  is typically of the form

  ```None
  -gp_likelihood(observations | p) + regularization(p)
  ```

  where regularization may be a negative prior log probability. The likelihood
  term is approximately proportional to the number of observations. As the
  number of observations changes over the course of a study, dividing the loss
  by this number helps ensure that loss values are roughly of the same order of
  magnitude, such that a constant learning rate may be used for gradient-based
  optimizers. Vizier library models make this adjustment automatically.

  Attributes:
    optimizer: Optax optimizer such as `optax.adam(1e-2)`.
    epochs: Number of train epochs.
    verbose: If >=1, logs the train progress. If >=2, logs the gradients.
  """

  optimizer: optax.GradientTransformation = attr.field()
  epochs: int = attr.field(kw_only=True)
  verbose: int = attr.field(kw_only=True, default=0, converter=int)

  # TODO: Prevent retracing.
  def __call__(
      self,
      init_params: core.Params,
      loss_fn: core.LossFunction[core.Params],
      rng: jax.Array,
      *,
      constraints: Optional[sp.Constraint] = None,
      best_n: Optional[int] = None,
  ) -> tuple[core.Params, chex.ArrayTree]:
    if constraints is None or constraints.bijector is None:
      bijector = None
      unconstrained_loss_fn = loss_fn
    else:
      bijector = constraints.bijector
      unconstrained_loss_fn = lambda x: loss_fn(bijector(x))

    grad_fn = jax.value_and_grad(unconstrained_loss_fn, has_aux=True)

    def _setup_all(init_params: core.Params) -> tuple[core.Params, OptState]:
      """Sets up both model params and optimizer state."""
      params = (
          bijector.inverse(init_params) if bijector is not None else init_params
      )
      opt_state = self.optimizer.init(params)
      return params, opt_state

    def _train_step(
        params: core.Params, opt_state: OptState
    ) -> tuple[core.Params, OptState, chex.ArrayTree]:
      """One train step."""
      (loss, metrics), grads = grad_fn(params)
      logging.log_if(logging.INFO, 'gradients: %s', self.verbose >= 2, grads)
      updates, opt_state = self.optimizer.update(grads, opt_state, params)
      params = optax.apply_updates(params, updates)
      metrics['loss'] = loss
      return params, opt_state, metrics

    params, opt_state = jax.vmap(_setup_all)(init_params)
    train_step = jax.vmap(_train_step)

    logging.info('Initialized parameters. %s',
                 jax.tree_map(lambda x: x.shape, params))

    # See https://jax.readthedocs.io/en/latest/faq.html#buffer-donation.
    train_step = jax.jit(train_step, donate_argnums=[0, 1])
    metrics = []
    for epoch in range(self.epochs):
      params, opt_state, step_metrics = train_step(params, opt_state)
      logging.log_if(
          logging.INFO,
          'Epoch %s: metrics: %s',
          self.verbose >= 1,
          epoch,
          step_metrics,
      )
      metrics.append(step_metrics)

    # Convert `metrics` from a list of dicts to a dict of arrays with leftmost
    # dimension corresponding to train steps.
    outer_treedef = jax.tree_util.tree_structure([0] * self.epochs)
    transposed_metrics = jax.tree_util.tree_transpose(
        outer_treedef, jax.tree_util.tree_structure(metrics[0]), metrics
    )
    metrics = jax.tree_util.tree_map(
        jnp.array,
        transposed_metrics,
        is_leaf=lambda x: jax.tree_util.tree_structure(x) == outer_treedef,
    )

    final_losses = metrics['loss'][-1]
    logging.info('Final loss: %s', final_losses)
    best_params = core.get_best_params(final_losses, params, best_n=best_n)
    if bijector is not None:
      best_params = bijector(best_params)
    return (best_params, metrics)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Utilities for handling GRPC API."""

from typing import Optional
import grpc
from vizier._src.service import custom_errors


# TODO: Create a type Union[_InactiveRpcError, LocalRpcError]
class LocalRpcError(grpc.RpcError):
  """Class for imitating both a `grpc.ServiceContext` and a `grpc.RpcError`."""

  _code: grpc.StatusCode = grpc.StatusCode.UNKNOWN
  _details: Optional[str] = None

  def set_code(self, code: grpc.StatusCode) -> None:
    self._code = code

  def set_details(self, details: str) -> None:
    self._details = details

  def code(self) -> grpc.StatusCode:
    return self._code

  def details(self) -> Optional[str]:
    return self._details


# TODO: Use this for all service errors.
def handle_exception(
    e: Exception, context: Optional[grpc.ServicerContext] = None
) -> None:
  """Converts custom exception into correct context error code.

  The rules for gRPC are:

  1) In the remote case (servicer wrapped into a server), the context is
  automatically generated by gRPC. Calling `context.set_code()` will
  automatically trigger an ` _InactiveRpcError` on the client side, which can
  collect the code and details from said RPC error.

  2) In the local case (e.g. when using a `VizierServicer` only), contexts are
  not used and set to None. In order to imitate the behavior of 1), we will
  instead use `LocalRpcError` to imitate both the behavior of a
  `ServicerContext` for setting codes and details, AND as an `_InactiveRpcError`
  to be raised on the client side.

  Args:
    e: Exception to be wrapped.
    context: For collecting error code and details. Set to None in the local
      case.

  Raises:
    LocalRpcError: If in the local case (when context is None).
  """
  if context is None:
    context = LocalRpcError(e)

  if isinstance(
      e, (custom_errors.ImmutableStudyError, custom_errors.ImmutableTrialError)
  ):
    context.set_code(grpc.StatusCode.FAILED_PRECONDITION)
  elif isinstance(e, custom_errors.NotFoundError):
    context.set_code(grpc.StatusCode.NOT_FOUND)
  elif isinstance(e, custom_errors.AlreadyExistsError):
    context.set_code(grpc.StatusCode.ALREADY_EXISTS)
  else:
    context.set_code(grpc.StatusCode.UNKNOWN)

  context.set_details(str(e))

  if isinstance(context, LocalRpcError):
    raise context



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for client."""

import abc
import functools
from typing import Callable, Optional

from absl import logging
from vizier import pyvizier as vz
from vizier.client import client_abc

from absl.testing import parameterized

# Aliases are defined, so when you are developing a new client, you can
# swap it with your subclass. It makes your IDE understand which class
# you are using.
_StudyClient = client_abc.StudyInterface
_TrialClient = client_abc.TrialInterface


class VizierClientTestMixin(abc.ABC):

  @abc.abstractmethod
  def create_study(
      self, study_config: vz.ProblemStatement, name: str
  ) -> _StudyClient:
    """Create study given study config and study name."""
    pass


class MyMeta(type(VizierClientTestMixin), type(parameterized.TestCase)):
  pass


class TestCase(parameterized.TestCase, VizierClientTestMixin, metaclass=MyMeta):
  """Generic tests for cross-platform clients.

  This test provides basic coverage and it is not meant to be a comprehensive
  e2e test on all possible inputs to the client.

  Override `create_study` method.
  """

  def create_test_study(self, name: str) -> _StudyClient:
    """Creates a study."""
    logging.info('Creating study name=%s, testcasename=%s', name, self.id())
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('float', 0.0, 1.0)
    problem.metric_information.append(
        vz.MetricInformation(
            name='maximize_metric', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    )
    study = self.create_study(problem, name)

    # TODO: Put this line back once we have __eq__ well-defined.
    # self.assertEqual(study.materialize_problem_statement(), problem)
    return study

  def test_create_or_load_study(self):
    study = self.create_test_study(self.id())
    study2 = study.from_resource_name(study.resource_name)
    self.assertEqual(study.resource_name, study2.resource_name)

    self.assertEqual(
        study.materialize_problem_statement(),
        study2.materialize_problem_statement(),
    )

  def test_delete_study(self):
    study = self.create_test_study(self.id())
    resource_name = study.resource_name
    study.delete()

    with self.assertRaises(Exception):
      study.from_resource_name(resource_name)

  def _example_trials(self) -> list[vz.Trial]:
    """Generates example trials."""
    trials = [
        # Completed trial.
        vz.Trial(
            parameters={'float': 0.5},
        ).complete(vz.Measurement({'maximize_metric': 1.0})),
        # Completed trial.
        vz.Trial(
            parameters={'float': 0.5},
        ).complete(vz.Measurement({'maximize_metric': 0.5})),
        # Requested trial, which will be made active below.
        vz.Trial(
            parameters={'float': 0.5},
            measurements=[vz.Measurement({'maximize_metric': 0.7})],
        ),
        # Requested trial.
        vz.Trial(parameters={'float': 0.5}, is_requested=True),
    ]
    for idx, t in enumerate(trials):
      t.metadata['future_id'] = str(idx + 1)  # id to be assigned
    return trials

  def create_test_study_with_trials(self, name: str) -> _StudyClient:
    study = self.create_test_study(name)
    trials = self._example_trials()
    for i, t in enumerate(trials):
      # TODO: Remove this.
      study._add_trial(t)  # pylint: disable=protected-access
      if i == 2:
        # Make sure the requested trial becomes ACTIVE.
        _ = study.suggest(count=1)
    return study

  @parameterized.parameters(list(state for state in vz.StudyState))
  def test_set_state(self, target_state: vz.StudyState):
    study = self.create_test_study_with_trials(self.id())
    # TODO: Check that the study moved to the target state.
    # This test is currently a placeholder.
    try:
      study.set_state(target_state)
    except NotImplementedError:
      logging.exception(
          'Set study state for %s is not implemented in %s',
          target_state,
          type(self),
      )

  def test_list_trials(self):
    study = self.create_test_study_with_trials(self.id())
    self.assertLen(list(study.trials()), 4)

  def test_trials_iter_and_get_are_equal(self):
    study = self.create_test_study_with_trials(self.id())
    all_trials = study.trials()
    self.assertEqual(
        [t.id for t in all_trials], [t.id for t in all_trials.get()]
    )

  def test_optimal_trials_on_empty(self):
    study = self.create_test_study(self.id())
    self.assertEmpty(list(study.optimal_trials()))

  def test_optimal_trials_notempty(self):
    study = self.create_test_study_with_trials(self.id())
    optimal_trials = list(study.optimal_trials())
    self.assertLen(optimal_trials, 1)
    self.assertEqual(
        optimal_trials[0].materialize().status, vz.TrialStatus.COMPLETED
    )

  def test_suggest_same_worker(self):
    # Given the same client id, should return the same trials.
    study = self.create_test_study(self.id())
    trials1 = study.suggest(count=5, client_id='worker1')
    self.assertLen(trials1, 5)
    trials2 = study.suggest(count=2, client_id='worker1')
    self.assertTrue({t.id for t in trials2}.issubset({t.id for t in trials1}))

  def test_suggest_requested(self):
    # Given the same client id, should return the same trials.
    study = self.create_test_study(self.id())
    request = vz.TrialSuggestion(parameters={'float': 0.11112})
    study.request(request)
    trials = study.suggest(count=5, client_id='worker1')
    self.assertLen(trials, 5)
    self.assertCountEqual(
        trials[0].parameters.items(), request.parameters.as_dict().items()
    )

  def test_suggest_different_workers(self):
    # Given different client ids, should generate new suggestions.
    study = self.create_test_study(self.id())
    worker1_trials = study.suggest(count=5, client_id='worker1')
    self.assertLen(worker1_trials, 5)
    worker1_trial_ids = {t.id for t in worker1_trials}
    worker2_trials = study.suggest(count=2, client_id='worker2')
    self.assertLen(worker2_trials, 2)
    worker2_trial_ids = {t.id for t in worker2_trials}
    self.assertEmpty(
        worker1_trial_ids.intersection(worker2_trial_ids),
        msg=(
            f'worker1_trial_ids={worker1_trial_ids}\n'
            f'worker2_trial_ids={worker2_trial_ids}'
        ),
    )

  def test_suggest_with_immutable_study(self):
    # Given immutable study, suggestions should be empty.
    study = self.create_test_study(self.id())
    study.set_state(vz.StudyState.ABORTED)
    trials = study.suggest(count=5, client_id='worker1')
    self.assertEmpty(trials)

  def test_get_trial(self):
    study = self.create_test_study_with_trials(self.id())
    trial = study.get_trial(2).materialize()
    self.assertEqual(trial.metadata['future_id'], '2')
    self.assertEqual(trial.id, 2)

  def test_get_trials(self):
    study = self.create_test_study_with_trials(self.id())
    completed = study.trials(vz.TrialFilter(status={vz.TrialStatus.COMPLETED}))
    self.assertLen(list(completed), 2)

  def test_parameters(self):
    study = self.create_test_study_with_trials(self.id())
    self.assertEqual(
        study.get_trial(1).parameters, study.get_trial(1).parameters
    )

  def test_study_update_metadata(self):
    """Checks for correct merge behavior."""
    study = self.create_test_study(self.id())
    delta_metadata = vz.Metadata({'bar': 'bar_v'}, foo='foo_v')
    study.update_metadata(delta_metadata)
    self.assertEqual(
        study.materialize_problem_statement().metadata.get('bar'), 'bar_v'
    )

    delta_metadata_2 = vz.Metadata({'bar': 'bar_w'})
    study.update_metadata(delta_metadata_2)

    problem_statement = study.materialize_problem_statement()
    self.assertEqual(problem_statement.metadata.get('bar'), 'bar_w')
    self.assertEqual(problem_statement.metadata.get('foo'), 'foo_v')

  ###########################################
  ## Tests on Trial class start here.
  ###########################################

  def test_delete_trial(self):
    study = self.create_test_study_with_trials(self.id())
    study.get_trial(2).delete()
    with self.assertRaises(Exception):
      study.get_trial(2)

  def test_complete_trial_no_measurements(self):
    study = self.create_test_study_with_trials(self.id())
    with self.assertRaises(Exception):
      study.get_trial(4).complete()

  def test_complete_trial_auto_selection(self):
    study = self.create_test_study_with_trials(self.id())
    trial = study.get_trial(3)
    measurement = trial.complete()
    self.assertIn(measurement, trial.materialize().measurements)

  def test_complete_trial_no_measurements_infeasible(self):
    study = self.create_test_study_with_trials(self.id())
    # Delete trial 3 so it's not suggested.
    trial = study.get_trial(3)
    trial.delete()
    # Ask Vizier to suggest the trial so it becomes ACTIVE.
    trial = study.suggest(count=1)[0]
    self.assertEqual(trial.id, 4)
    self.assertIsNone(trial.complete(infeasible_reason='just because'))
    self.assertTrue(trial.materialize().infeasible)

  def test_complete_trial_manual_measurement(self):
    study = self.create_test_study_with_trials(self.id())
    trial = study.get_trial(3)
    final_measurement = vz.Measurement(metrics={'maximize_metric': 0.1})
    self.assertEqual(trial.complete(final_measurement), final_measurement)
    self.assertEqual(trial.materialize().status, vz.TrialStatus.COMPLETED)

  def test_should_trial_stop(self):
    # TODO: Improve coverage.
    study = self.create_test_study_with_trials(self.id())
    trial = study.get_trial(3)
    self.assertIsInstance(trial.check_early_stopping(), bool)

  def test_trial_stop(self):
    """Checks for correct stopping behavior."""
    study = self.create_test_study_with_trials(self.id())
    active_trial = study.suggest(count=1)[0]
    active_trial.stop()
    self.assertEqual(active_trial.materialize().status, vz.TrialStatus.STOPPING)

    # COMPLETED, STOPPING
    noop_trials = [study.get_trial(2), active_trial]
    original_statuses = [trial.materialize().status for trial in noop_trials]
    for trial, status in zip(noop_trials, original_statuses):
      trial.stop()  # Should do nothing.
      self.assertEqual(trial.materialize().status, status)

    requested_trial = study.get_trial(4)
    with self.assertRaises(Exception):
      requested_trial.stop()

  def test_intermediate_measurement(self):
    study = self.create_test_study_with_trials(self.id())
    trial = study.get_trial(3)
    before = trial.materialize()
    trial.add_measurement(
        vz.Measurement(steps=2, metrics={'maximize_metric': 0.2})
    )
    after = trial.materialize()
    self.assertLen(after.measurements, len(before.measurements) + 1)

  def test_intermediate_measurement_infeasible(self):
    study = self.create_test_study_with_trials(self.id())
    trial = study.get_trial(3)
    trial.complete(infeasible_reason='just because')
    before = trial.materialize()
    trial.add_measurement(
        vz.Measurement(steps=2, metrics={'maximize_metric': 0.2})
    )
    after = trial.materialize()
    # Can't add measurements to infeasible trials, but doing so isn't an error.
    self.assertEqual(before, after)

  def test_intermediate_measurement_completed(self):
    study = self.create_test_study_with_trials(self.id())
    trial = study.get_trial(3)
    trial.complete(vz.Measurement(steps=1, metrics={'maximize_metric': 0.5}))
    # Can't add measurements to completed trials; doing so is an error.
    with self.assertRaises(Exception):
      trial.add_measurement(
          vz.Measurement(steps=2, metrics={'maximize_metric': 0.2})
      )

  def test_study_property(self):
    study = self.create_test_study_with_trials(self.id())
    trial = study.get_trial(2)
    self.assertEqual(
        trial.materialize(), trial.study.get_trial(2).materialize()
    )

  def assertPassesE2ETuning(
      self,
      *,
      study_factory: Optional[
          Callable[[vz.ProblemStatement], _StudyClient]
      ] = None,
      batch_size: int = 2,
      num_iterations: int = 5,
      multi_objective: bool = False,
  ):
    """Runs an e2e test.

    This test simulates a hyperparameter tuning scenario, where the model
    hyperparameters are being tuned, and automated trial early stopping is
    used to stop trials early.

    Args:
      study_factory: If not specified, uses the default factory of this class.
      batch_size:
      num_iterations:
      multi_objective:
    """
    study_factory = study_factory or functools.partial(
        self.create_study, study_id=self.id()
    )

    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param(
        'learning_rate', min_value=0.0, max_value=1.0, default_value=0.5
    )
    problem.search_space.root.add_int_param(
        'num_layers', min_value=1, max_value=5
    )
    problem.metric_information = [
        vz.MetricInformation(
            name='accuracy', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    ]
    if multi_objective:
      problem.metric_information.append(
          vz.MetricInformation(
              name='latency', goal=vz.ObjectiveMetricGoal.MINIMIZE
          )
      )

    study: _StudyClient = study_factory(problem)

    def learning_curve_simulator(learning_rate: float) -> list[float]:
      return [learning_rate * step for step in range(10)]

    for _ in range(num_iterations):
      suggestions = study.suggest(count=batch_size)
      for trial in suggestions:
        trial: _TrialClient = trial
        assert isinstance(trial, _TrialClient), type(trial)
        learning_rate = trial.parameters['learning_rate']
        num_layers = trial.parameters['num_layers']
        possible_curve = learning_curve_simulator(learning_rate)
        evaluated_curve = []
        for i, obj in enumerate(possible_curve, start=1):
          if i > 1 and trial.check_early_stopping():
            break
          evaluated_curve.append(obj)
          trial.add_measurement(
              vz.Measurement(
                  steps=i,
                  elapsed_secs=0.1 * i,
                  metrics={'accuracy': obj, 'latency': 0.5 * num_layers},
              )
          )
        trial.complete()
        stored_curve = [
            m.metrics['accuracy'].value
            for m in trial.materialize().measurements
        ]

        # See if curve was stored correctly.
        self.assertEqual(evaluated_curve, stored_curve)

        # See if final_measurement is defaulted to end of curve.
        final_measurement = trial.materialize().final_measurement
        assert final_measurement is not None
        final_accuracy = final_measurement.metrics['accuracy'].value
        self.assertEqual(evaluated_curve[-1], final_accuracy)

  def test_trial_update_metadata(self):
    """Checks for correct merge behavior."""
    study = self.create_test_study_with_trials(self.id())
    trial = study.get_trial(3)
    delta_metadata = vz.Metadata({'bar': 'bar_v'}, foo='foo_v')
    trial.update_metadata(delta_metadata)
    self.assertEqual(trial.materialize().metadata.get('bar'), 'bar_v')
    delta_metadata_2 = vz.Metadata({'bar': 'bar_w'})
    trial.update_metadata(delta_metadata_2)
    self.assertEqual(trial.materialize().metadata.get('bar'), 'bar_w')
    self.assertEqual(trial.materialize().metadata.get('foo'), 'foo_v')



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import numpy as np
from vizier import pyvizier as vz
from vizier._src.pythia import policy as pythia
from vizier._src.pythia import suggest_default

from absl.testing import absltest


space = vz.SearchSpace()
space.root.add_int_param('x', 0, 1000)

empty_study_descriptor = vz.StudyDescriptor(
    vz.ProblemStatement(space),
    guid='1',
    max_trial_id=0,
)


class GetDefaultParametersTest(absltest.TestCase):

  def test_double_user_default(self):
    ss = vz.SearchSpace()
    ss.root.add_float_param('x', 0.0, 1.0, default_value=0.2)
    params = suggest_default._get_default_parameters(ss)
    self.assertEqual(params.get_value('x'), 0.2)

  @unittest.skip('TODO')
  def test_double_logscale(self):
    ss = vz.SearchSpace()
    ss.root.add_float_param(
        'x', np.exp(-2), np.exp(2), scale_type=vz.ScaleType.LOG
    )
    params = suggest_default._get_default_parameters(ss)
    self.assertEqual(params.get_value('x'), 1.0)

  def test_double_fixed(self):
    ss = vz.SearchSpace()
    ss.root.add_float_param('x', 1.0, 1.0)
    params = suggest_default._get_default_parameters(ss)
    self.assertEqual(params.get_value('x'), 1.0)

  def test_discrete(self):
    ss = vz.SearchSpace()
    ss.root.add_discrete_param('x', [1, 2, 3, 6])
    params = suggest_default._get_default_parameters(ss)
    self.assertEqual(params.get_value('x'), 3)

  def test_categorical(self):
    ss = vz.SearchSpace()
    ss.root.add_categorical_param('x', ['a', 'b', 'c', 'd'])
    params = suggest_default._get_default_parameters(ss)
    self.assertEqual(params.get_value('x'), 'c')


class SuggestDefaultTest(absltest.TestCase):

  def test_decorator(self):
    class FakeDecoratedPolicy(pythia.Policy):

      @suggest_default.seed_with_default
      def suggest(
          self, request: pythia.SuggestRequest
      ) -> pythia.SuggestDecision:
        if request.count != 2:
          raise ValueError(f'count must be 2. Was: {request.count}')
        if request._study_descriptor != empty_study_descriptor:
          raise ValueError('study_descriptor must match request')
        return pythia.SuggestDecision(
            [vz.TrialSuggestion({'x': 999}), vz.TrialSuggestion({'x': 1000})]
        )

      def early_stop(
          self, request: pythia.EarlyStopRequest
      ) -> pythia.EarlyStopDecisions:
        return pythia.EarlyStopDecisions()

    FakeDecoratedPolicy().suggest(
        pythia.SuggestRequest(count=3, study_descriptor=empty_study_descriptor)
    )

  def test_init_override(self):
    class FakeInitOverridePolicy(pythia.Policy):

      def __init__(self):
        self.suggest = suggest_default.seed_with_default(self.suggest)

      def suggest(
          self, request: pythia.SuggestRequest
      ) -> pythia.SuggestDecision:
        if request.count != 2:
          raise ValueError(f'count must be 2. Was: {request.count}')
        if request._study_descriptor != empty_study_descriptor:
          raise ValueError('study_descriptor must match request')
        return pythia.SuggestDecision(
            [vz.TrialSuggestion({'x': 999}), vz.TrialSuggestion({'x': 1000})]
        )

      def early_stop(
          self, request: pythia.EarlyStopRequest
      ) -> pythia.EarlyStopDecisions:
        return pythia.EarlyStopDecisions()

    FakeInitOverridePolicy().suggest(
        pythia.SuggestRequest(count=3, study_descriptor=empty_study_descriptor)
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import numpy as np
from vizier import pyvizier
from vizier._src.benchmarks.experimenters import discretizing_experimenter
from vizier._src.benchmarks.experimenters import numpy_experimenter
from vizier._src.benchmarks.experimenters import permuting_experimenter
from vizier._src.benchmarks.experimenters.synthetic import bbob

from absl.testing import absltest
from absl.testing import parameterized


class PermutingExperimenterTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('Sphere', bbob.Sphere),
      ('Rastrigin', bbob.Rastrigin),
      ('BuecheRastrigin', bbob.BuecheRastrigin),
      ('LinearSlope', bbob.LinearSlope),
      ('AttractiveSector', bbob.AttractiveSector),
      ('StepEllipsoidal', bbob.StepEllipsoidal),
      ('RosenbrockRotated', bbob.RosenbrockRotated),
      ('Discus', bbob.Discus),
      ('BentCigar', bbob.BentCigar),
      ('SharpRidge', bbob.SharpRidge),
      ('DifferentPowers', bbob.DifferentPowers),
      ('Weierstrass', bbob.Weierstrass),
      ('SchaffersF7', bbob.SchaffersF7),
      ('SchaffersF7IllConditioned', bbob.SchaffersF7IllConditioned),
      ('GriewankRosenbrock', bbob.GriewankRosenbrock),
      ('Schwefel', bbob.Schwefel),
      ('Katsuura', bbob.Katsuura),
      ('Lunacek', bbob.Lunacek),
      ('Gallagher101Me', bbob.Gallagher101Me),
  )
  def testNumpyExperimenter(self, func):
    dim = 3
    exptr = numpy_experimenter.NumpyExperimenter(
        func, bbob.DefaultBBOBProblemStatement(dim)
    )

    # Asserts parameters are the same.
    parameters = list(exptr.problem_statement().search_space.parameters)
    self.assertLen(parameters, dim)

    discretization = {
        parameters[0].name: ['0.1', '0.5', '-0.1'],
        parameters[1].name: [0.2, 1.2, 2.2],
    }

    dis_exptr = discretizing_experimenter.DiscretizingExperimenter(
        exptr, discretization
    )
    discretized_parameters = [
        p.name
        for p in dis_exptr.problem_statement().search_space.parameters
        if np.isfinite(p.num_feasible_values)
    ]
    self.assertLen(discretized_parameters, 2)
    permuted_exptr = permuting_experimenter.PermutingExperimenter(
        dis_exptr, discretized_parameters
    )

    self.assertEqual(
        permuted_exptr.problem_statement(), dis_exptr.problem_statement()
    )

    parameters = {
        parameters[0].name: '-0.1',
        parameters[1].name: 1.2,
        parameters[2].name: 1.5,
    }
    t = pyvizier.Trial(parameters=parameters)

    permuted_exptr.evaluate([t])
    metric_name = exptr.problem_statement().metric_information.item().name
    self.assertNotAlmostEqual(
        func(np.array([-0.1, 1.2, 1.5])),
        t.final_measurement_or_die.metrics[metric_name].value,
    )
    self.assertEqual(t.status, pyvizier.TrialStatus.COMPLETED)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Decorators for Policy.suggest."""

import functools
import types
from typing import TypeVar

import attrs
from vizier import pyvizier as vz
from vizier._src.pythia.policy import Policy
from vizier._src.pythia.policy import SuggestDecision
from vizier._src.pythia.policy import SuggestRequest


_T = TypeVar('_T')


def _get_default_parameters(search_space: vz.SearchSpace) -> vz.ParameterDict:
  """Gets the default parameters for the given search space."""
  builder = vz.SequentialParameterBuilder(search_space)

  for pc in builder:
    if pc.default_value is not None:
      builder.choose_value(pc.default_value)
    elif pc.type in (
        vz.ParameterType.CATEGORICAL,
        vz.ParameterType.INTEGER,
        vz.ParameterType.DISCRETE,
    ):
      # Choose the middle value.
      builder.choose_value(pc.feasible_values[len(pc.feasible_values) // 2])
    elif pc.type == vz.ParameterType.DOUBLE:
      if pc.num_feasible_values == 1:
        builder.choose_value(pc.bounds[0])
      else:
        # TODO: Handle scaling properly.
        midpoint = (pc.bounds[0] + pc.bounds[1]) / 2
        builder.choose_value(midpoint)
  return builder.parameters


def seed_with_default(suggest_fn: _T) -> _T:
  """Decorator for Policy.suggest to always suggest the default or center.

  How to use as a decorator:

  ```
  class MyPolicy(Policy):
    @seed_with_default
    def suggest(self, ...):
      ...
  ```

  How to use as a function:
  class MyPolicy(Policy):

    def __init__(self, ..., *, use_seed_with_default: bool):
      if use_seed_with_default:
        self.suggest = seed_with_default(self.suggest)

    def suggest(self, ...):
      ...

  Args:
    suggest_fn:

  Returns:
    suggest_fn that suggest the default or center of the search space if
    the study is empty.
  """

  if hasattr(suggest_fn, '__self__'):
    unbound = seed_with_default(suggest_fn.__func__)
    return types.MethodType(unbound, suggest_fn.__self__)

  @functools.wraps(suggest_fn)
  def wrapper_fn(self: Policy, request: SuggestRequest) -> SuggestDecision:
    """If study is empty, suggests a default trial before using the policy."""
    if request.max_trial_id > 0:
      return suggest_fn(self, request)

    default_parameters = _get_default_parameters(
        request.study_config.search_space
    )
    decision = SuggestDecision([vz.TrialSuggestion(default_parameters)])

    if request.count > 1:
      more_suggestions = suggest_fn(
          self, attrs.evolve(request, count=request.count - 1)
      )
      decision.suggestions.extend(more_suggestions.suggestions)
      decision.metadata = more_suggestions.metadata

    return decision

  return wrapper_fn
